#ifndef FUTURE_EXTENSIONS_H
#define FUTURE_EXTENSIONS_H

#include <glib-object.h>
#include <telepathy-glib/channel.h>
#include <telepathy-glib/connection.h>

#include "tests/lib/glib/future/extensions/_gen/enums.h"
#include "tests/lib/glib/future/extensions/_gen/svc-channel.h"
#include "tests/lib/glib/future/extensions/_gen/svc-connection.h"
#include "tests/lib/glib/future/extensions/_gen/svc-misc.h"

G_BEGIN_DECLS

#include "tests/lib/glib/future/extensions/_gen/gtypes.h"
#include "tests/lib/glib/future/extensions/_gen/interfaces.h"

G_END_DECLS

#endif
