/*
 * conn.h - header for a connection that implements Conn.I.Addressing
 *
 * Copyright (C) 2011 Collabora Ltd. <http://www.collabora.co.uk/>
 * Copyright (C) 2011 Nokia Corporation
 *
 * Copying and distribution of this file, with or without modification,
 * are permitted in any medium without royalty provided the copyright
 * notice and this notice are preserved.
 */

#ifndef __TP_TESTS_ADDRESSING_CONN_H__
#define __TP_TESTS_ADDRESSING_CONN_H__

#include <glib-object.h>
#include <telepathy-glib/base-connection.h>

#include "contacts-conn.h"

G_BEGIN_DECLS

typedef struct _TpTestsAddressingConnection TpTestsAddressingConnection;
typedef struct _TpTestsAddressingConnectionClass TpTestsAddressingConnectionClass;
typedef struct _TpTestsAddressingConnectionPrivate TpTestsAddressingConnectionPrivate;

struct _TpTestsAddressingConnectionClass {
    TpTestsContactsConnectionClass parent_class;
};

struct _TpTestsAddressingConnection {
    TpTestsContactsConnection parent;

    TpTestsAddressingConnectionPrivate *priv;
};

GType tp_tests_addressing_connection_get_type (void);

/* TYPE MACROS */
#define TP_TESTS_TYPE_ADDRESSING_CONNECTION \
  (tp_tests_addressing_connection_get_type ())
#define TP_TESTS_ADDRESSING_CONNECTION(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), TP_TESTS_TYPE_ADDRESSING_CONNECTION, \
                              TpTestsAddressingConnection))
#define TP_TESTS_ADDRESSING_CONNECTION_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass), TP_TESTS_TYPE_ADDRESSING_CONNECTION, \
                           TpTestsAddressingConnectionClass))
#define TP_TESTS_ADDRESSING_IS_CONNECTION(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), TP_TESTS_TYPE_ADDRESSING_CONNECTION))
#define TP_TESTS_ADDRESSING_IS_CONNECTION_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass), TP_TESTS_TYPE_ADDRESSING_CONNECTION))
#define TP_TESTS_ADDRESSING_CONNECTION_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS ((obj), TP_TESTS_TYPE_ADDRESSING_CONNECTION, \
                              TpTestsAddressingConnectionClass))

G_END_DECLS

#endif
